from datetime import datetime
from typing import TYPE_CHECKING, Any, Type

import ultima_scraper_api
from sqlalchemy import ScalarResult, Select, func, or_, select
from sqlalchemy.ext.asyncio.session import async_object_session
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import and_, exists, select, union_all
from ultima_scraper_api.apis.onlyfans.classes.comment_model import (
    CommentModel as OFCommentModel,
)
from ultima_scraper_api.apis.onlyfans.classes.user_model import (
    create_user as OFUserModel,
)
from ultima_scraper_collection.helpers.main_helper import is_valuable
from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
    ContentMetadata,
)
from ultima_scraper_db.databases.ultima.schemas.templates.site import (
    BoughtContentModel,
    CommentModel,
    FilePathModel,
    MediaModel,
    MessageModel,
    NotificationModel,
    PostModel,
    StoryModel,
    SubscriptionModel,
    UserAuthModel,
    UserInfoModel,
    UserModel,
)
from ultima_scraper_db.managers.database_manager import Schema

content_model_types = StoryModel | PostModel | MessageModel
from ultima_scraper_api import post_types

if TYPE_CHECKING:
    from ultima_scraper_collection import datascraper_types


class ContentManager:
    def __init__(
        self,
        user: "UserModel",
    ):
        self.__user__ = user

    async def init(self):
        await self.__user__.awaitable_attrs._stories
        await self.__user__.awaitable_attrs._posts
        await self.__user__.awaitable_attrs._messages
        self.stories: list["StoryModel"] = self.__user__._stories  # type: ignore
        self.posts: list["PostModel"] = self.__user__._posts  # type: ignore
        self.messages: list["MessageModel"] = self.__user__._messages  # type: ignore
        return self

    async def get_contents(self, content_type: str | None = None):
        if content_type:
            # if not hasattr(self, content_type.lower()):
            #     empty_list: content_model_types = []
            #     return empty_list
            result: list[content_model_types] = getattr(self, content_type.lower())
            return result
        return self.stories + self.posts + self.messages

    async def add_content(self, content: ContentMetadata):
        match content.api_type:
            case "Stories":
                content_model = StoryModel(
                    id=content.content_id,
                    user_id=content.user_id,
                    created_at=content.__soft__.created_at,
                )
                self.stories.append(content_model)
            case "Posts":
                content_model = PostModel(
                    id=content.content_id,
                    user_id=content.user_id,
                    text=content.text,
                    price=content.price,
                    paid=int(content.paid),
                    created_at=content.__soft__.created_at,
                )
                self.posts.append(content_model)
            case "Messages":
                content_model = MessageModel(
                    id=content.content_id,
                    user_id=content.user_id,
                    receiver_id=content.receiver_id,
                    text=content.text,
                    price=content.price,
                    paid=int(content.paid),
                    verified=True,
                    created_at=content.__soft__.created_at,
                )
                self.messages.append(content_model)
                session = async_object_session(self.__user__)
                assert session
                await session.flush()
                assert content.receiver_id
                content_model.receiver_id = content.receiver_id
                pass
            case _:
                raise Exception("Content not assigned")
        return content_model

    async def find_content(
        self,
        content_id: int,
    ):
        temp_contents = await self.get_contents()
        for content in temp_contents:
            if content.id == content_id:
                return content

    async def find_paid_contents(self):
        from sqlalchemy.ext.asyncio import async_object_session

        session = async_object_session(self.__user__)
        assert session
        stmt1 = (
            select(MediaModel)
            .join(PostModel.media)
            .options(joinedload(MediaModel.filepaths))
            .where(
                and_(
                    PostModel.user_id == self.__user__.id,
                    PostModel.price > 0,
                    PostModel.paid == True,
                    exists(MediaModel).where(MediaModel.filepaths.any()),
                    MediaModel.preview == False,
                )
            )
        )
        stmt2 = (
            select(MediaModel)
            .join(MessageModel.media)
            .options(joinedload(MediaModel.filepaths))
            .where(
                and_(
                    MessageModel.user_id == self.__user__.id,
                    MessageModel.price > 0,
                    MessageModel.paid == True,
                    exists(MediaModel).where(MediaModel.filepaths.any()),
                    MediaModel.preview == False,
                )
            )
        )
        union_stmt = union_all(stmt1, stmt2)
        orm_stmt = select(MediaModel).from_statement(union_stmt)
        result_3 = await session.scalars(orm_stmt)
        result_3 = result_3.unique().all()
        pass
        return result_3

    async def sum(self):
        session = async_object_session(self.__user__)
        assert session

        stmt1 = (
            select(func.sum(MediaModel.size))
            .join(StoryModel.media)
            .where(StoryModel.user_id == self.__user__.id)
        )
        stmt2 = (
            select(func.sum(MediaModel.size))
            .join(PostModel.media)
            .where(PostModel.user_id == self.__user__.id)
        )
        stmt3 = (
            select(func.sum(MediaModel.size))
            .join(MessageModel.media)
            .where(MessageModel.user_id == self.__user__.id)
        )
        result1 = await session.execute(stmt1)
        result2 = await session.execute(stmt2)
        result3 = await session.execute(stmt3)
        final_sum = sum(
            (result1.scalar() or 0, result2.scalar() or 0, result3.scalar() or 0)
        )
        return final_sum

    async def media_sum(self, category: str):
        session = async_object_session(self.__user__)
        assert session

        stmt1 = (
            select(func.count())
            .where(MediaModel.category == category)
            .where(MediaModel.filepaths.any())
            .join(StoryModel.media)
            .where(StoryModel.user_id == self.__user__.id)
        )
        stmt2 = (
            select(func.count())
            .where(MediaModel.category == category)
            .where(MediaModel.filepaths.any())
            .join(PostModel.media)
            .where(PostModel.user_id == self.__user__.id)
        )
        stmt3 = (
            select(func.count())
            .where(MediaModel.category == category)
            .where(MediaModel.filepaths.any())
            .join(MessageModel.media)
            .where(MessageModel.user_id == self.__user__.id)
        )
        result1 = await session.scalar(stmt1)
        result2 = await session.scalar(stmt2)
        result3 = await session.scalar(stmt3)
        final_sum = sum((result1 or 0, result2 or 0, result3 or 0))
        return final_sum


class StatementBuilder:
    def __init__(self, model: Type[UserModel | UserInfoModel]) -> None:
        self.model = model
        self.statement: Select[Any] = select(model)

    def filter_by_user_identifiers(self, identifiers: list[int | str]):
        final_identifiers = [str(x) for x in identifiers]
        template_query = [
            UserModel.username.in_([x for x in final_identifiers if not x.isdigit()]),
            UserModel.id.in_([int(x) for x in final_identifiers if x.isdigit()]),
        ]
        if self.model == UserModel:
            self.statement = self.statement.filter(or_(*template_query))
        else:
            self.statement = self.statement.join(UserModel).filter(or_(*template_query))

        return self.statement

    def filter_by_description(self, description: str):
        description = description.lower()
        if self.model == UserModel:
            self.statement = self.statement.join(UserInfoModel).filter(
                UserInfoModel.description.ilike(f"%{description}%")
            )
        else:
            self.statement = self.statement.join(UserModel).filter(
                UserInfoModel.description.ilike(f"%{description}%")
            )
        return self.statement


class SiteDB:
    def __init__(
        self, schema: Schema, datascraper: "datascraper_types | None" = None
    ) -> None:
        self.database = schema.database
        self.schema = self.database.get_schema(schema.name)
        self.session = self.schema.session
        self.datascraper = datascraper

    async def get_users(self, identifiers: list[int | str]):
        stmt_builder = StatementBuilder(UserModel).filter_by_user_identifiers(
            identifiers
        )
        db_users: ScalarResult[UserModel] = await self.session.scalars(stmt_builder)
        return db_users.all()

    async def get_user(self, identifier: int | str):
        stmt_builder = StatementBuilder(UserModel).filter_by_user_identifiers(
            [identifier]
        )
        db_users: ScalarResult[UserModel] = await self.session.scalars(stmt_builder)
        return db_users.first()

    async def get_media(self, media_id: int):
        stmnt = select(MediaModel).where(MediaModel.id == media_id)
        found_media = await self.session.scalars(stmnt)
        return found_media.first()

    async def update_user(
        self, api_user: ultima_scraper_api.user_types, found_db_user: UserModel | None
    ):
        _db_user = await self.create_or_update_user(
            api_user, existing_user=found_db_user
        )
        await self.session.commit()

        current_job = api_user.get_current_job()
        if current_job:
            assert current_job
            current_job.done = True

    async def create_or_update_user(
        self, api_user: ultima_scraper_api.user_types, existing_user: UserModel | None
    ):
        session = self.schema.session
        db_user = existing_user or UserModel()
        if not existing_user:
            db_user.id = api_user.id
            session.add(db_user)
        found_user_info = await db_user.find_user_info(api_user.username)
        if api_user.username != db_user.username:
            checkpoint_ui = await db_user.find_user_info(db_user.username)
            if not found_user_info:
                user_info = await self.create_or_update_user_info(
                    api_user, checkpoint_ui=checkpoint_ui
                )
                await db_user.awaitable_attrs.user_infos
                db_user.user_infos.append(user_info)
                pass
            else:
                user_info = await self.create_or_update_user_info(
                    api_user, existing_user_info=found_user_info
                )
                pass
        else:
            user_info = await self.create_or_update_user_info(
                api_user, existing_user_info=found_user_info
            )
            pass
        if not existing_user:
            status = False
            if await is_valuable(api_user):
                if await api_user.subscription_price() == 0:
                    if (
                        isinstance(api_user, OFUserModel)
                        and await api_user.get_paid_contents()
                    ):
                        status = True
                else:
                    status = True
            if status:
                notification = NotificationModel(
                    user_id=api_user.id, category="new_performer"
                )
                session.add(notification)
            await session.flush()
        if existing_user:
            _alias = await db_user.add_alias(api_user.username)
        if api_user.is_authed_user():
            api_authed = api_user.get_authed()
            await self.create_or_update_auth_info(api_authed, db_user)
            if api_authed.is_authed():
                await db_user.activate()

            paid_contents = await api_authed.get_paid_content()
            for paid_content in paid_contents:
                if isinstance(paid_content, dict):
                    continue
                supplier = paid_content.get_author()
                temp_local_user = await session.scalars(
                    select(UserModel).where(UserModel.id == supplier.id)
                )
                local_user = temp_local_user.first()
                await self.create_or_update_user(supplier, local_user)
                found_bought_content = await db_user.find_bought_content(supplier.id)
                if not found_bought_content:
                    bought_content = BoughtContentModel(supplier_id=supplier.id)
                    db_user.bought_contents.append(bought_content)
                    pass
                else:
                    pass
            pass
            api_subscriptions = await api_authed.get_subscriptions()
            for api_subscription in api_subscriptions:
                await self.create_or_update_subscription(api_subscription, db_user)
                pass
            pass
        pass
        if isinstance(api_user, OFUserModel):
            socials = await api_user.get_socials()
            await db_user.add_socials(socials)

            spotify = await api_user.get_spotify()
            if spotify:
                spotify["socialMedia"] = "spotify"
                spotify["username"] = spotify["displayName"]
                await db_user.add_socials([spotify])
        db_user.username = api_user.username
        db_user.balance = api_user.creditBalance
        db_user.performer = api_user.isPerformer
        db_user.join_date = (
            datetime.fromisoformat(api_user.joinDate) if api_user.joinDate else None
        )
        _s = await db_user.awaitable_attrs._stories
        _p = await db_user.awaitable_attrs._posts
        _m = await db_user.awaitable_attrs._messages
        pass
        for _key, contents in api_user.content_manager.categorized.__dict__.items():
            for _, content in contents.items():
                await self.create_or_update_content(db_user, content)
        db_user.last_checked_at = datetime.now()
        user_info.size = await db_user.content_manager.sum()
        await self.session.flush()
        return db_user

    async def create_or_update_auth_info(
        self, api_authed: ultima_scraper_api.auth_types, db_user: UserModel
    ):
        await db_user.awaitable_attrs.user_auth_info
        db_user_auth_info = db_user.user_auth_info
        if not db_user_auth_info:
            exported_auth_details = api_authed.get_auth_details().export(UserAuthModel)
            user_auth_model = UserAuthModel(**exported_auth_details)
            user_auth_model.active = user_auth_model.active
            db_user.user_auth_info = user_auth_model
            pass
        else:
            exported_auth_details = api_authed.get_auth_details().export()
            db_user.user_auth_info.update(exported_auth_details)
            pass
        db_user.user_auth_info.email = api_authed.user.email
        return db_user.user_auth_info

    async def create_or_update_subscription(
        self,
        subscription: ultima_scraper_api.subscription_types,
        db_authed: "UserModel",
    ):
        temp_db_sub_user = await self.session.scalars(
            select(UserModel).where(UserModel.id == subscription.user.id)
        )
        db_sub_user = temp_db_sub_user.first()
        if db_sub_user:
            pass
        db_sub_user = await self.create_or_update_user(subscription.user, db_sub_user)
        subscription_user = subscription.user
        db_subscription = await db_authed.find_subscription(subscription_user.id)
        authed = subscription_user.get_authed()
        if not db_subscription:
            db_subscription = SubscriptionModel()
            await db_sub_user.awaitable_attrs.subscribers
            db_sub_user.subscribers.append(db_subscription)

        # Common lines of code for both if and else cases
        db_subscription.user_id = subscription_user.id
        db_subscription.subscriber_id = authed.id
        if isinstance(subscription_user, OFUserModel):
            db_subscription.paid_content = bool(
                await subscription_user.get_paid_contents()
            )
        pass
        expires_at = subscription.resolve_expires_at()
        db_subscription.expires_at = expires_at
        db_subscription.active = subscription.is_active()
        await self.session.flush()
        await db_sub_user.awaitable_attrs.subscribers
        return db_subscription

    async def create_or_update_user_info(
        self,
        subscription_user: ultima_scraper_api.user_types,
        existing_user_info: "UserInfoModel | None" = None,
        checkpoint_ui: "UserInfoModel | None" = None,
    ):
        user_info = existing_user_info or UserInfoModel()
        user_info.username = subscription_user.username
        user_info.name = subscription_user.name
        user_info.description = subscription_user.about
        user_info.post_count = subscription_user.postsCount
        user_info.image_count = subscription_user.photosCount
        user_info.video_count = subscription_user.videosCount
        user_info.audio_count = subscription_user.audiosCount
        user_info.favourited_count = subscription_user.favoritedCount
        pass
        user_info.price = await subscription_user.subscription_price() or 0
        from ultima_scraper_api.apis.onlyfans.classes.user_model import create_user

        user_info.promotion = (
            bool(await subscription_user.get_promotions())
            if isinstance(subscription_user, create_user)
            else False
        )
        user_info.location = subscription_user.location
        user_info.website = subscription_user.website
        if not existing_user_info and checkpoint_ui:
            user_info.sex = checkpoint_ui.sex
            user_info.size = checkpoint_ui.size
            user_info.downloaded_at = checkpoint_ui.downloaded_at
            user_info.uploaded_at = checkpoint_ui.uploaded_at
        await self.session.flush()
        return user_info

    async def create_or_update_content(
        self, db_performer: UserModel, content: ContentMetadata
    ):
        api_performer = content.__soft__.get_author()
        api_type = content.api_type
        content_manager = await db_performer.content_manager.init()
        found_db_content = await content_manager.find_content(content.content_id)
        if not found_db_content:
            pass
        if content.paid:
            pass
        db_content = found_db_content or await content_manager.add_content(content)
        db_content_paid = None
        content.paid = False
        if isinstance(db_content, PostModel | MessageModel):
            if not db_content.paid:
                db_content.paid = True if content.paid else False
                if (
                    bool(db_content.paid) == False
                    and content.price
                    and content.price > 0
                ):
                    assert self.datascraper
                    fmu = self.datascraper.filesystem_manager.get_file_manager(
                        api_performer.id
                    )
                    valid_local_media_count = 0
                    for file in fmu.files:
                        remote_media_date = content.__soft__.created_at
                        remote_post_date = content.__soft__.created_at
                        stat = file.stat()
                        mt = stat.st_mtime
                        local_media_date = datetime.fromtimestamp(mt).replace(
                            microsecond=0
                        )
                        local_post_date = datetime.fromtimestamp(mt).replace(
                            microsecond=0
                        )
                        if (
                            local_media_date == remote_media_date
                            or local_post_date == remote_post_date
                        ):
                            valid_local_media_count += 1
                    if valid_local_media_count == content.__soft__.mediaCount:
                        db_content.paid = True
            db_content_paid = db_content.paid
            db_content.price = content.price or 0
        if content.preview_media_ids:
            if "poll" not in content.preview_media_ids:
                pass
            pass
        if isinstance(db_content, MessageModel):
            db_content.verified = True
        db_content.created_at = content.created_at

        for media in content.medias:
            assert media.id
            final_url = media.urls[0] if media.urls else None
            found_db_media = await db_content.find_media(media.id)
            if found_db_media:
                db_media = found_db_media
                if not final_url:
                    final_url = db_media.url
                    if final_url:
                        # UNFINISHED
                        # Need to find a way to rebuild media's directory, filename, etc (We can use the reformatter)
                        # content_metadata = ContentMetadata(db_content.id, api_type)
                        # extractor = DBContentExtractor(db_content)
                        # extractor.__api__ = api
                        # await content_metadata.resolve_extractor(extractor)
                        # Remove none when solution found
                        final_url = None
                        pass
            else:
                found_global_media = await self.get_media(media.id)
                if found_global_media:
                    db_media = found_global_media
                else:
                    db_media = MediaModel(
                        id=media.id,
                        url=final_url,
                        size=0,
                        preview=media.preview,
                        created_at=media.created_at,
                    )
                await db_content.awaitable_attrs.media
                db_content.media.append(db_media)
            if not final_url:
                pass
            if db_media.created_at and db_media.created_at.tzinfo is None:
                db_media.created_at = db_media.created_at
            filepath = await db_media.find_filepath(db_content.id, api_type)
            if not filepath and final_url:
                assert media.directory
                assert media.filename
                filepath = FilePathModel(
                    # media_id=db_media.id,
                    filepath=media.directory.joinpath(media.filename).as_posix(),
                    preview=media.preview,
                )
                await filepath.set_content(db_content)
                await db_media.awaitable_attrs.filepaths
                db_media.filepaths.append(filepath)
            if final_url:
                if db_content_paid or (not content.paid and not content.price):
                    # No previews get through here
                    db_media.preview = False
                    db_media.url = final_url
                    assert filepath and media.directory and media.filename
                    filepath.media_id = db_media.id
                    filepath.filepath = media.directory.joinpath(
                        media.filename
                    ).as_posix()
                    filepath.preview = False
                else:
                    if isinstance(db_content, PostModel | MessageModel):
                        # Handle previews
                        if not db_content.paid and not db_media.url:
                            db_media.preview = True
                            db_media.url = final_url
                            assert filepath and media.directory and media.filename
                            filepath.media_id = db_media.id
                            filepath.filepath = media.directory.joinpath(
                                media.filename
                            ).as_posix()
                            filepath.preview = True
            elif db_content_paid:
                db_media.preview = False
                if filepath:
                    filepath.preview = False
            db_media.category = media.media_type
            if media.size >= int(db_media.size):
                db_media.size = media.size
            media_created_at = media.created_at
            if db_media.created_at is None or media_created_at < db_media.created_at:
                if db_media.created_at is None:
                    pass
                db_media.created_at = media_created_at
        await self.session.flush()
        if isinstance(content.__soft__, post_types):
            if len(content.__soft__.comments) > 1:
                pass
            for comment in content.__soft__.comments:
                found_db_comment = await self.session.scalar(
                    select(CommentModel).where(CommentModel.id == comment.id)
                )
                if not found_db_comment:
                    giphy_id = None
                    reply_id = None
                    if isinstance(comment, OFCommentModel):
                        giphy_id = comment.giphy_id
                    else:
                        if comment.reply_id != comment.reply_root_id:
                            reply_id = comment.reply_id
                    db_comment = CommentModel(
                        id=comment.id,
                        post_id=db_content.id,
                        reply_id=reply_id,
                        user_id=db_content.user_id,
                        giphy_id=giphy_id,
                        text=comment.text,
                        likes_count=comment.likes_count,
                        created_at=comment.created_at,
                    )
                    self.session.add(db_comment)
                else:
                    pass
        pass



from datetime import datetime
from typing import TYPE_CHECKING, Any

from sqlalchemy import (
    BigInteger,
    Boolean,
    Float,
    ForeignKey,
    Integer,
    SmallInteger,
    String,
    Text,
    UniqueConstraint,
    and_,
    select,
)
from sqlalchemy.ext.asyncio import async_object_session
from sqlalchemy.ext.compiler import compiles  # type: ignore
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ultima_scraper_db.databases.ultima import (
    CustomFuncs,
    DefaultContentTypes,
    SiteTemplate,
)
from ultima_scraper_db.helpers import TIMESTAMPTZ, selectin_relationship

if TYPE_CHECKING:
    from ultima_scraper_db.databases.ultima.schemas.management import SiteModel


class UserModel(SiteTemplate):
    __tablename__ = "users"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True, autoincrement=False)
    username: Mapped[str] = mapped_column(Text, nullable=True)
    balance: Mapped[float] = mapped_column(Float, server_default="0")
    spend: Mapped[bool] = mapped_column(Boolean, server_default="false")
    performer: Mapped[bool] = mapped_column(Boolean, server_default="false")
    favorite: Mapped[bool] = mapped_column(Boolean, server_default="false")
    active: Mapped[bool] = mapped_column(Boolean, server_default="true")
    downloaded_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ, nullable=True)
    last_checked_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ, nullable=True)
    join_date: Mapped[datetime | None] = mapped_column(TIMESTAMPTZ, nullable=True)
    created_at: Mapped[datetime] = mapped_column(
        TIMESTAMPTZ, server_default=CustomFuncs.utcnow()
    )
    updated_at: Mapped[datetime] = mapped_column(
        TIMESTAMPTZ, onupdate=CustomFuncs.utcnow(), nullable=True
    )
    user_auth_info: Mapped["UserAuthModel"] = selectin_relationship()
    user_infos: Mapped[list["UserInfoModel"]] = relationship(
        foreign_keys="UserInfoModel.user_id", back_populates="user"
    )

    subscriptions: Mapped[list["SubscriptionModel"]] = relationship(
        "SubscriptionModel",
        foreign_keys="SubscriptionModel.subscriber_id",
        back_populates="subscriber",
    )

    subscribers: Mapped[list["SubscriptionModel"]] = relationship(
        "SubscriptionModel",
        foreign_keys="SubscriptionModel.user_id",
    )

    socials: Mapped[list["SocialModel"]] = relationship(back_populates="user")
    _stories: Mapped[list["StoryModel"]] = relationship()
    _posts: Mapped[list["PostModel"]] = relationship()
    _messages: Mapped[list["MessageModel"]] = relationship(
        primaryjoin="or_(UserModel.id==MessageModel.user_id, "
        "UserModel.id==MessageModel.receiver_id)",
    )
    aliases: Mapped[list["UserAliasModel"]] = relationship(back_populates="user")
    jobs: Mapped[list["JobModel"]] = relationship(back_populates="user")

    bought_contents: Mapped[list["BoughtContentModel"]] = relationship(
        foreign_keys="BoughtContentModel.buyer_id"
    )
    supplied_contents: Mapped[list["BoughtContentModel"]] = relationship(
        foreign_keys="BoughtContentModel.supplier_id"
    )

    @property
    def content_manager(self):
        from ultima_scraper_db.managers.site_db import ContentManager

        return ContentManager(self)

    async def find_content(self):
        pass

    async def add_alias(self, username: str):
        if self.username != username:
            alias = await self.find_aliases(self.username)
            if not alias:
                alias = UserAliasModel(username=self.username)
                self.aliases.append(alias)
            self.username = username
            return alias

    async def find_aliases(self, username: str):
        await self.awaitable_attrs.aliases
        for alias in self.aliases:
            if alias.username == username:
                return alias

    async def find_subscription(self, user_id: int):
        await self.awaitable_attrs.subscriptions
        for subscription in self.subscriptions:
            if subscription.user_id == user_id:
                return subscription

    async def has_active_subscription(self):
        await self.awaitable_attrs.subscribers
        return bool(
            [x for x in self.subscribers if datetime.now().astimezone() < x.expires_at]
        )

    async def find_user_info(self, username: str):
        await self.awaitable_attrs.user_infos
        for user_info in self.user_infos:
            if user_info.username == username:
                return user_info

    async def add_socials(self, socials: list[dict[str, Any]]):
        for social in socials:
            site_name: str = social["socialMedia"]
            identifier = social["username"]
            found_social = await self.find_socials(site_name)
            if not found_social:
                if not identifier:
                    match site_name.lower():
                        case "amazon" | "etsy" | "discord":
                            identifier = social["link"]
                        case "instagram":
                            identifier = social["url"]
                        case _:
                            raise Exception("social identifier not found")
                    pass
                db_social = SocialModel(site=site_name, identifier=identifier)
                self.socials.append(db_social)
                pass

    async def find_socials(self, site_name: str, identifier: str | None = None):
        await self.awaitable_attrs.socials
        for social in self.socials:
            if social.site == site_name:
                if social.identifier == identifier:
                    return social
                if identifier is None:
                    return social

    async def add_job(
        self, server_id: int, category: str, site: "SiteModel", skippable: bool = False
    ):
        # Need to create algo to resolve server_id by how many videos user has
        await self.awaitable_attrs.jobs
        job = JobModel(
            site_id=site.id,
            user_username=self.username,
            category=category,
            server_id=server_id,
            skippable=skippable,
        )
        self.jobs.append(job)
        return job

    async def find_job(self, category: str):
        found_job = None
        await self.awaitable_attrs.jobs
        for job in self.jobs:
            if job.category == category:
                found_job = job
        return found_job

    async def find_bought_content(self, performer_id: int):
        await self.awaitable_attrs.bought_contents
        for bought_content in self.bought_contents:
            if bought_content.supplier_id == performer_id:
                return bought_content

    async def get_supplied_content(self, active: bool = False):
        final_supplied_content: list[BoughtContentModel] = []
        await self.awaitable_attrs.supplied_contents
        if active:
            for supplied_content in self.supplied_contents:
                await supplied_content.awaitable_attrs.buyer
                if not supplied_content.buyer.user_auth_info.active:
                    continue
                final_supplied_content.append(supplied_content)
        else:
            final_supplied_content = self.supplied_contents
        return final_supplied_content

    async def find_buyers(self, active: bool = False):
        temp_buyers: set[UserModel] = set()
        for content in await self.get_supplied_content(active=active):
            temp_buyers.add(await content.awaitable_attrs.buyer)
        await self.awaitable_attrs.subscribers
        session = async_object_session(self)
        assert session
        query = (
            select(UserModel)
            .join(SubscriptionModel.subscriber)
            .where(SubscriptionModel.user_id == self.id)
        )
        if active:
            query = query.where(SubscriptionModel.expires_at >= datetime.now()).where(
                UserModel.user_auth_info.has(active=True)
            )
        subscribers = await session.scalars(query)
        for subscriber in subscribers:
            temp_buyers.add(subscriber)
        final_buyers = temp_buyers
        return list(final_buyers)

    async def activate(self):
        self.user_auth_info.active = True
        await self.awaitable_attrs.subscriptions
        for subscription in self.subscriptions:
            await subscription.awaitable_attrs.user
            if subscription.user.active:
                subscription.active = True
            else:
                pass
        pass

    async def deactivate(self):
        await self.awaitable_attrs.subscriptions
        self.user_auth_info.active = False
        for subscription in self.subscriptions:
            subscription.active = False


class UserAuthModel(SiteTemplate):
    __tablename__ = "user_auths"
    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    cookie: Mapped[str] = mapped_column(Text, nullable=True)
    x_bc: Mapped[str] = mapped_column(Text, nullable=True)
    authorization: Mapped[str] = mapped_column(Text, nullable=True)
    user_agent: Mapped[str] = mapped_column(Text)
    email: Mapped[str] = mapped_column(Text, nullable=True)
    password: Mapped[str] = mapped_column(Text, nullable=True)
    active: Mapped[bool] = mapped_column(Boolean, server_default="true")

    def update(self, info: dict[str, Any]):
        for k, v in info.items():
            if k == "id":
                continue
            setattr(self, k, v)


class UserInfoModel(SiteTemplate):
    __tablename__ = "user_infos"

    __table_args__ = (UniqueConstraint("user_id", "username"),)
    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    username: Mapped[str] = mapped_column(Text, nullable=True)
    name: Mapped[str] = mapped_column(Text, nullable=True)
    price: Mapped[int] = mapped_column(Float, server_default="0")
    description: Mapped[str] = mapped_column(Text, nullable=True)
    promotion: Mapped[bool] = mapped_column(
        Boolean, server_default="true", default="true"
    )
    sex: Mapped[int] = mapped_column(SmallInteger, nullable=True)
    post_count: Mapped[int] = mapped_column(Integer, server_default="0")
    image_count: Mapped[int] = mapped_column(Integer, server_default="0")
    video_count: Mapped[int] = mapped_column(Integer, server_default="0")
    audio_count: Mapped[int] = mapped_column(Integer, server_default="0")
    favourited_count: Mapped[int] = mapped_column(Integer, server_default="0")
    size: Mapped[int] = mapped_column(BigInteger, server_default="0")
    location: Mapped[str] = mapped_column(Text, nullable=True)
    website: Mapped[str] = mapped_column(Text, nullable=True)
    downloaded_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ, nullable=True)
    uploaded_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ, nullable=True)
    user: Mapped["UserModel"] = relationship(back_populates="user_infos")


class ContentMediaAssoModel(DefaultContentTypes, SiteTemplate):
    __tablename__ = "content_media_asso"
    __table_args__ = (
        UniqueConstraint("story_id", "post_id", "message_id", "media_id"),
    )
    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    media: Mapped["MediaModel"] = relationship(
        "MediaModel", back_populates="content_media_assos"
    )

    def get_key(self, match_value: str):
        from sqlalchemy import inspect

        for key in inspect(ContentMediaAssoModel).columns.keys():
            if match_value.lower() in key.lower():
                return key
        pass

    def get_user(self):
        from sqlalchemy import inspect

        exclusions = ["media"]

        for key, _ in inspect(ContentMediaAssoModel).relationships.items():
            if any(x for x in exclusions if x == key):
                continue
            final_value = getattr(self, key)
            if final_value:
                user: UserModel = final_value.user
                return user
        raise Exception("User not set")


class MediaModel(SiteTemplate):
    __tablename__ = "media"
    id: Mapped[int] = mapped_column(BigInteger, primary_key=True, autoincrement=False)
    url: Mapped[str] = mapped_column(Text, nullable=True)
    size: Mapped[int] = mapped_column(BigInteger, server_default="0", default="0")
    category: Mapped[str] = mapped_column(String, nullable=True)
    preview: Mapped[bool] = mapped_column(Boolean, server_default="false")
    created_at: Mapped[datetime | None] = mapped_column(TIMESTAMPTZ)
    filepaths: Mapped[list["FilePathModel"]] = relationship(
        "FilePathModel",
        # back_populates="media",
    )

    content_media_assos: Mapped[list[ContentMediaAssoModel]] = relationship(
        back_populates="media"
    )

    async def find_filepath(self, content_id: int, content_type: str):
        content_type = content_type if "Stories" != content_type else "Story"
        await self.awaitable_attrs.filepaths
        for filepath in self.filepaths:
            fp_content_id = getattr(
                filepath, f"{content_type.lower().removesuffix('s')}_id"
            )
            if fp_content_id == content_id:
                return filepath


class ContentTemplate(SiteTemplate):
    __abstract__ = True


class StoryModel(ContentTemplate):
    __tablename__ = "x_stories"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True, autoincrement=False)
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    created_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ)

    user: Mapped["UserModel"] = relationship(
        foreign_keys=user_id,
    )

    media: Mapped[list["MediaModel"]] = relationship(
        "MediaModel",
        secondary=ContentMediaAssoModel.__table__,
        primaryjoin=and_(
            ContentMediaAssoModel.story_id == id,
        ),
        secondaryjoin=ContentMediaAssoModel.media_id == MediaModel.id,
        backref="stories",
    )

    async def find_media(self, media_id: int):
        await self.awaitable_attrs.media
        for media in self.media:
            if media.id == media_id:
                return media


class PostModel(ContentTemplate):
    __tablename__ = "x_posts"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True, autoincrement=False)
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    text: Mapped[str] = mapped_column(Text, nullable=True)
    price: Mapped[float] = mapped_column(Float, server_default="0")
    deleted: Mapped[bool] = mapped_column(Boolean, server_default="false")
    archived: Mapped[bool] = mapped_column(Boolean, server_default="false")
    paid: Mapped[bool | None] = mapped_column(Boolean, nullable=True)
    created_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ)

    user: Mapped["UserModel"] = relationship(
        foreign_keys=user_id,
    )

    media: Mapped[list["MediaModel"]] = relationship(
        "MediaModel",
        secondary=ContentMediaAssoModel.__table__,
        primaryjoin=and_(
            ContentMediaAssoModel.post_id == id,
        ),
        secondaryjoin=ContentMediaAssoModel.media_id == MediaModel.id,
        backref="posts",
    )

    async def find_media(self, media_id: int):
        await self.awaitable_attrs.media
        for media in self.media:
            if media.id == media_id:
                return media


class MessageModel(ContentTemplate):
    __tablename__ = "x_messages"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True, autoincrement=False)
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    receiver_id: Mapped[int] = mapped_column(
        BigInteger, ForeignKey("users.id"), nullable=True
    )
    text: Mapped[str] = mapped_column(Text, nullable=True)
    price: Mapped[float] = mapped_column(Float, server_default="0")
    deleted: Mapped[bool] = mapped_column(Boolean, server_default="false")
    paid: Mapped[bool | None] = mapped_column(Boolean, nullable=True)
    verified: Mapped[bool] = mapped_column(Boolean, server_default="false")
    created_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ)

    user: Mapped["UserModel"] = relationship(
        foreign_keys=user_id,
    )
    media: Mapped[list["MediaModel"]] = relationship(
        "MediaModel",
        secondary=ContentMediaAssoModel.__table__,
        primaryjoin=and_(
            ContentMediaAssoModel.message_id == id,
        ),
        secondaryjoin=ContentMediaAssoModel.media_id == MediaModel.id,
        backref="messages",
    )

    async def find_media(self, media_id: int):
        await self.awaitable_attrs.media
        for media in self.media:
            if media.id == media_id:
                return media


class CommentModel(SiteTemplate):
    __tablename__ = "comments"
    id: Mapped[int] = mapped_column(BigInteger, primary_key=True, autoincrement=False)
    post_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("x_posts.id"))
    reply_id: Mapped[int] = mapped_column(
        BigInteger, ForeignKey("comments.id"), nullable=True
    )
    user_id: Mapped[int] = mapped_column(BigInteger)
    giphy_id: Mapped[str | None] = mapped_column(String, server_default=None)
    text: Mapped[str | None] = mapped_column(String, server_default=None)
    likes_count: Mapped[int] = mapped_column(Integer, server_default="0")
    created_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ)


class SubscriptionModel(SiteTemplate):
    __tablename__ = "subscriptions"
    __table_args__ = (UniqueConstraint("user_id", "subscriber_id"),)

    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    subscriber_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    paid_content: Mapped[bool] = mapped_column(Boolean, server_default="0")
    active: Mapped[bool] = mapped_column(Boolean, server_default="true")
    downloaded_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ, nullable=True)
    expires_at: Mapped[datetime] = mapped_column(TIMESTAMPTZ)

    user: Mapped["UserModel"] = relationship(
        "UserModel",
        foreign_keys="SubscriptionModel.user_id",
    )
    subscriber: Mapped["UserModel"] = relationship(
        "UserModel",
        back_populates="subscribers",
        foreign_keys="SubscriptionModel.subscriber_id",
    )


class FilePathModel(DefaultContentTypes, SiteTemplate):
    __tablename__ = "filepaths"
    __table_args__ = (
        UniqueConstraint("story_id", "post_id", "message_id", "media_id"),
    )

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    filepath: Mapped[str] = mapped_column(String(255))

    preview: Mapped[bool] = mapped_column(Boolean, server_default="false")
    downloaded: Mapped[bool] = mapped_column(Boolean, server_default="false")
    # media: Mapped["MediaModel"] = relationship(
    #     "MediaModel",
    #     # back_populates="filepaths",
    # )


class UserAliasModel(SiteTemplate):
    __tablename__ = "user_aliases"
    __table_args__ = (UniqueConstraint("user_id", "username"),)

    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    username: Mapped[str] = mapped_column(String(50))
    user: Mapped["UserModel"] = relationship(back_populates="aliases")


from sqlalchemy import event
from sqlalchemy.orm.attributes import AttributeEventToken


@event.listens_for(SubscriptionModel.downloaded_at, "set", propagate=True)
def update_downloaded_at(
    subscription: SubscriptionModel,
    value: datetime,
    oldvalue: datetime,
    initiator: AttributeEventToken,
):
    if value != oldvalue:
        subscription.user.downloaded_at = value


class JobModel(SiteTemplate):
    __tablename__ = "jobs"
    __table_args__ = (UniqueConstraint("site_id", "user_id", "category"),)

    id: Mapped[int] = mapped_column(primary_key=True)
    site_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("management.sites.id"))
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    user_username: Mapped[str] = mapped_column(Text)
    category: Mapped[str] = mapped_column(Text)
    server_id: Mapped[int] = mapped_column(
        BigInteger, ForeignKey("management.servers.id")
    )
    skippable: Mapped[bool] = mapped_column(Boolean, server_default="false")
    active: Mapped[bool] = mapped_column(Boolean, server_default="true")
    completed_at: Mapped[datetime | None] = mapped_column(TIMESTAMPTZ, nullable=True)
    user: Mapped[UserModel] = relationship(back_populates="jobs")
    site: Mapped["SiteModel"] = relationship(foreign_keys=site_id)


class NotificationModel(SiteTemplate):
    __tablename__ = "notifications"
    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(
        BigInteger, ForeignKey("users.id"), nullable=True
    )
    category: Mapped[str] = mapped_column(String)
    sent_discord: Mapped[bool] = mapped_column(Boolean, server_default="false")
    sent_telegram: Mapped[bool] = mapped_column(Boolean, server_default="false")
    created_at: Mapped[datetime] = mapped_column(
        TIMESTAMPTZ, server_default=CustomFuncs.utcnow()
    )

    user: Mapped["UserModel"] = selectin_relationship()


class SocialModel(SiteTemplate):
    __tablename__ = "socials"
    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    site: Mapped[str] = mapped_column(String)
    identifier: Mapped[str] = mapped_column(String)
    user: Mapped[UserModel] = relationship(back_populates="socials")


class BoughtContentModel(SiteTemplate):
    __tablename__ = "bought_content"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    supplier_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))
    buyer_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("users.id"))

    supplier: Mapped[UserModel] = relationship(foreign_keys=supplier_id)
    buyer: Mapped[UserModel] = relationship(foreign_keys=buyer_id)



from sqlalchemy import BigInteger, Boolean, SmallInteger, Text, orm, select
from sqlalchemy.orm import Mapped, mapped_column

from ultima_scraper_db.databases.ultima import ManagementTemplate
from ultima_scraper_db.databases.ultima.schemas.templates.site import (
    JobModel,
    UserModel,
)
from ultima_scraper_db.managers.database_manager import Schema


class ServerModel(ManagementTemplate):
    __tablename__ = "servers"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(Text)
    ip: Mapped[str] = mapped_column(Text)
    mac_address: Mapped[str] = mapped_column(Text)
    job_limit: Mapped[int] = mapped_column(SmallInteger, server_default="10")
    active: Mapped[int] = mapped_column(SmallInteger, server_default="1")


class SiteModel(ManagementTemplate):
    __tablename__ = "sites"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(Text)
    db_name: Mapped[str] = mapped_column(Text)
    url: Mapped[str] = mapped_column(Text)
    active: Mapped[bool] = mapped_column(Boolean, server_default="true")
    size: Mapped[int] = mapped_column(BigInteger, server_default="0")
    user_id_checkpoint: Mapped[int] = mapped_column(BigInteger, nullable=True)

    async def get_jobs(self, site_db: Schema, server_id: int | None = None):
        stmt = select(JobModel).filter_by(site_id=self.id)
        if server_id:
            stmt = stmt.filter_by(server_id=server_id)

        stmt = (
            stmt.join(JobModel.user)
            .order_by(JobModel.id.asc())
            .order_by(UserModel.downloaded_at.desc())
            .options(orm.contains_eager(JobModel.user))
        )
        result = await site_db.session.scalars(stmt)
        jobs = result.all()
        return jobs


default_sites = [
    SiteModel(name="OnlyFans", db_name="onlyfans", url="https://onlyfans.com"),
    SiteModel(name="Fansly", db_name="fansly", url="https://fansly.com"),
]



import itertools
from pathlib import Path
from typing import Any

from alembic import command
from alembic.config import Config
from alembic.migration import MigrationContext
from alembic.operations.ops import MigrationScript
from alembic.util import CommandError  # type: ignore
from sqlalchemy import Connection, MetaData, inspect
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sshtunnel import SSHTunnelForwarder  # type: ignore

from ultima_scraper_db.helpers import create_database, database_exists


class Alembica:
    def __init__(
        self, database_path: str | Path | None = None, generate: bool = False
    ) -> None:
        if not database_path:
            self.database_path = Path(__file__).parent.parent.joinpath(
                "databases/ultima"
            )
        else:
            self.database_path = Path(database_path)
        assert self.database_path.exists(), "Invalid database directory"
        self.migration_directory = self.database_path.joinpath("alembic")
        self.config = Config(self.database_path.joinpath("alembic.ini"))
        self.config.set_main_option(
            "script_location",
            self.migration_directory.as_posix(),
        )
        self.is_generate = generate


class SSHConnection:
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        private_key: bytes,
        private_key_password: str | None = None,
    ) -> None:
        self.host = host
        self.port = port
        self.username = username
        self.private_key = private_key
        self.password = private_key_password


class Schema:
    def __init__(
        self,
        name: str,
        engine: AsyncEngine,
        session: AsyncSession,
        database: "Database",
    ) -> None:
        self.name = name
        self.engine = engine
        self.session = session
        self.sessionmaker = async_sessionmaker(engine, expire_on_commit=False)
        self.database = database


def show_schemas(conn: Connection):
    inspector = inspect(conn)
    schemas = inspector.get_schema_names()
    return schemas


class Database:
    def __init__(
        self,
        name: str,
        host: str,
        port: int,
        username: str,
        password: str,
        ssh: SSHTunnelForwarder | None,
        metadata: MetaData | None = None,
        alembica: Alembica | None = None,
    ) -> None:
        self.name = name
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        if ssh:
            ssh.start()
        self.ssh: SSHTunnelForwarder | None = ssh
        self.schemas: dict[str, Schema] = {}
        self.metadata = metadata
        self.alembica = alembica

    async def init_db(self):
        if self.ssh:
            connection_string = f"postgresql+asyncpg://{self.username}:{self.password}@{self.ssh.local_bind_host}:{self.ssh.local_bind_port}/{self.name}"  # type: ignore
        else:
            connection_string = f"postgresql+asyncpg://{self.username}:{self.password}@{self.host}:{self.port}/{self.name}"
        engine = create_async_engine(
            connection_string,
            echo=True,
        )
        self.engine = engine

        if not await database_exists(connection_string):
            await self.setup()
        from sqlalchemy.schema import CreateSchema

        async with engine.connect() as conn:

            def create_schemas(session: Connection):
                if not self.metadata:
                    return
                for schema_name in self.metadata._schemas:  # type: ignore
                    session.execute(CreateSchema(schema_name, if_not_exists=True))
                session.commit()

            await conn.run_sync(create_schemas)
            await self.resolve_schemas()
        return self

    async def resolve_schemas(self):
        connection_string = str(self.engine.url.render_as_string(hide_password=False))
        async with self.engine.connect() as conn:
            schema_strings: list[str] = await conn.run_sync(show_schemas)
            assert schema_strings
            for schema_string in schema_strings:
                engine = create_async_engine(
                    connection_string,
                    echo=True,
                    execution_options={"schema_translate_map": {None: schema_string}},
                    pool_size=20,
                )
                async_session = AsyncSession(engine, expire_on_commit=False)
                schema_obj = Schema(schema_string, engine, async_session, self)

                self.schemas[schema_obj.name] = schema_obj
            return self.schemas

    async def create(self):
        await create_database(self.engine.url.render_as_string(hide_password=False))

    def get_schema(self, name: str):
        return self.schemas[name.lower()]

    async def setup(self):
        await self.create()

    async def generate_migration(self):
        def get_current_revision(connection: Connection):
            context = MigrationContext.configure(connection)
            current_rev = context.get_current_revision()
            return current_rev

        def process_revision_directives(
            context: MigrationContext,
            revision: tuple[str, str],
            directives: list[Any],
        ):
            revisions: list[MigrationScript] = directives
            # Prevent actually generating a migration
            has_upgrades = list(
                itertools.chain.from_iterable(
                    op.as_diffs()
                    for script in revisions
                    for op in script.upgrade_ops_list
                )
            )
            if not has_upgrades:
                directives[:] = []

        def run_revision(
            connection: Connection,
            cfg: Config,
            autogenerate: bool,
            local_metadata: MetaData,
        ):
            cfg.attributes["connection"] = connection  # type:ignore
            revision = command.revision(
                alembic_cfg,
                autogenerate=autogenerate,
                process_revision_directives=process_revision_directives,
            )
            return revision

        async with self.engine.connect() as conn:
            current_rev = await conn.run_sync(get_current_revision)
            assert self.alembica
            alembic_cfg = self.alembica.config
            if not current_rev:
                try:
                    _revision = await conn.run_sync(
                        run_revision, alembic_cfg, True, self.metadata
                    )
                    pass
                except CommandError as _e:
                    await self.run_migrations()
                    pass
            else:
                _revision = await conn.run_sync(
                    run_revision, alembic_cfg, True, self.metadata
                )
                pass
            pass

    async def run_migrations(self) -> None:
        def run_upgrade(connection: Connection, cfg: Config):
            cfg.attributes["connection"] = connection  # type:ignore
            command.upgrade(alembic_cfg, "head")
            return True

        while True:
            try:
                async with self.engine.connect() as conn:
                    assert self.alembica
                    alembic_cfg = self.alembica.config
                    _upgraded = await conn.run_sync(run_upgrade, alembic_cfg)
                    break
            except Exception as e:
                print(e)
                pass


class DatabaseManager:
    def __init__(self) -> None:
        self.databases: dict[str, Database] = {}

    def create_database(
        self,
        name: str,
        host: str,
        port: int,
        username: str,
        password: str,
        ssh: SSHTunnelForwarder,
        metadata: MetaData | None = None,
        alembica: Alembica | None = None,
    ):
        database = Database(
            name, host, port, username, password, ssh, metadata, alembica
        )
        return database

    def add_database(self, database: Database):
        self.databases[database.name] = database
        return database

    def resolve_database(self, name: str):
        return self.databases[name]
